// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/async/samples/rxasync/main.c $
// $Rev: 53128 $
// $Date: 2023-06-16 15:09:50 -0500 (Fri, 16 Jun 2023) $

// SIO4: Asynchronous Protocol Library: Sample Application: source file

#include "main.h"



//*****************************************************************************
static int _parse_args(int argc, char** argv, args_t* args)
{
	const gsc_arg_item_t	list[]	=
	{
		//	type				var						arg			values									desc
		{ GSC_ARG_S32_FLAG,		&args->xcvr,			"-232",		{ SIO4_XCVR_PROTOCOL_RS232 },			"Select RS-232 transceivers, if possible."						},
		{ GSC_ARG_S32_FLAG,		&args->xcvr,			"-422",		{ SIO4_XCVR_PROTOCOL_RS422_RS485 },		"Select RS-422/485 transceivers, if possible (default)."		},
		{ GSC_ARG_S32_FLAG,		&args->xcvr,			"-423",		{ SIO4_XCVR_PROTOCOL_RS423 },			"Select RS-423 transceivers, if possible."						},
		{ GSC_ARG_S32_FLAG,		&args->continuous,		"-c",		{ -1 },									"Continue testing until an error occurs."						},
		{ GSC_ARG_S32_FLAG,		&args->continuous,		"-C",		{ +1 },									"Continue testing even if errors occur."						},
		{ GSC_ARG_S32_MIN,		&args->minute_limit,	"-m",		{ 1 },									"Continue testing for up to # minutes."							},
		{ GSC_ARG_S32_MIN,		&args->test_limit,		"-n",		{ 1 },									"Continue testing for up to # iterations."						},
		{ GSC_ARG_S32_FLAG,		&args->print,			"-p",		{ 1 },									"Print received data to the screen."							},
		{ GSC_ARG_S32_FLAG,		&args->pc,				"-pc",		{ 1 },									"Configure the channel to receive data from a PC serial port."	},
		{ GSC_ARG_S32_FLAG,		&args->save,			"-s",		{ 1 },									"Save received data to file 'data.txt'."						},
		{ GSC_ARG_S32_RANGE,	&args->seconds,			"-s",		{ 1, 3600 },							"Read in data for # seconds (default is 15)."					},
		{ GSC_ARG_S32_FLAG,		&args->term_res,		"-t",		{ SIO4_ASYNC_CABLE_TERM_ENABLE_YES },	"Enable the termination resistors."								},
		{ GSC_ARG_S32_FLAG,		&args->verbose,			"-v",		{ 1 },									"Show all settings before reading data."						},
		{ GSC_ARG_DEV_INDEX,	&args->index,			"index",	{ 1 },									"The index of the device to access."							}
	};

	const gsc_arg_set_t	set	=
	{
		/* name	*/	"rxasync",
		/* desc	*/	"Perform Asynchronous data reception.",
		/* qty	*/	SIZEOF_ARRAY(list),
		/* list	*/	list
	};

	int	errs;

	memset(args, 0, sizeof(args_t));
	args->seconds	= 15;
	args->term_res	= SIO4_ASYNC_CABLE_TERM_ENABLE_NO;
	args->xcvr		= SIO4_XCVR_PROTOCOL_RS422_RS485;

	errs	= gsc_args_parse(argc, argv, &set);

	return(errs);
}



//*****************************************************************************
static void _show_access_index(const args_t* args)
{
	gsc_label("Accessing Device");
	printf("%d\n", args->index);
}



//*****************************************************************************
static void _show_time_stamp(void)
{
	const char*	psz;
	struct tm*	stm;
	time_t		tt;

	time(&tt);
	stm	= localtime(&tt);
	psz	= (char*) asctime(stm);
	gsc_label("Time Stamp");
	printf("%s", psz);
}



/******************************************************************************
*
*	Function:	main
*
*	Purpose:
*
*		Control the overall flow of the application.
*
*	Arguments:
*
*		argc			The number of command line arguments.
*
*		argv			The list of command line arguments.
*
*	Returned:
*
*		EXIT_SUCCESS	No errors were encounterred.
*		EXIT_FAILURE	One or more problems were encounterred.
*
******************************************************************************/

int main(int argc, char** argv)
{
	args_t	args;
	int		errs;
	time_t	exec		= time(NULL);
	long	failures	= 0;
	long	hours;
	long	mins;
	time_t	now;
	long	passes		= 0;
	int		ret;
	long	secs;
	time_t	t_limit;
	time_t	test;
	long	tests		= 0;

	for (;;)	// A convenience loop.
	{
		gsc_label_init(26);
		test	= time(NULL);
		errs	= _parse_args(argc, argv, &args);

		// Introduction

		t_limit	= exec + (args.minute_limit * 60);
		os_id_host();
		_show_time_stamp();
		_show_access_index(&args);
		errs	+= async_init_util(1);
		errs	+= os_id_driver(sio4_open, sio4_read, sio4_close);
		errs	+= sio4_count_boards(1, &args.qty);

		if (args.qty <= 0)
			errs++;

		// Open access to device

		if (errs == 0)
			errs	= async_open_util(args.index, 0, -1, 1, &args.fd);

		if (errs)
			args.continuous	= 0;

		if (errs == 0)
		{
			// Device identification

			errs	= sio4_id_device(args.fd, -1, 1);

			// Perform testing

			errs	+= perform_tests(&args, &args.skip);

			// Close device access

			errs	+= async_close_util(args.fd, -1, 1);
		}

		// End of processing.

		now	= time(NULL);
		tests++;

		if (errs)
		{
			failures++;
			printf(	"\nRESULTS: FAIL <---  (%d error%s)",
					errs,
					(errs == 1) ? "" : "s");
		}
		else if (args.skip)
		{
			printf("\nRESULTS: SKIPPED");
		}
		else
		{
			passes++;
			printf("\nRESULTS: PASS");
		}

		secs	= (long) (now - test);
		hours	= secs / 3600;
		secs	= secs % 3600;
		mins	= secs / 60;
		secs	= secs % 60;
		printf(" (duration %ld:%ld:%02ld)\n", hours, mins, secs);

		secs	= (long) (now - exec);
		hours	= secs / 3600;
		secs	= secs % 3600;
		mins	= secs / 60;
		secs	= secs % 60;
		printf(	"SUMMARY: tests %ld, pass %ld, fail %ld"
				" (duration %ld:%ld:%02ld)\n\n",
				tests,
				passes,
				failures,
				hours,
				mins,
				secs);

		if (args.skip)
			break;

		if (args.continuous == 0)
			break;

		if ((args.continuous < 0) && (errs))
			break;

		if ((args.minute_limit > 0) && (now >= t_limit))
			break;

		if ((args.test_limit > 0) && (tests >= args.test_limit))
			break;
	}

	ret	= ((args.qty <= 0) || (errs) || (failures)) ? EXIT_FAILURE : EXIT_SUCCESS;
	return(ret);
}


