// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/async/samples/txasync/perform.c $
// $Rev: 54294 $
// $Date: 2024-03-19 10:36:16 -0500 (Tue, 19 Mar 2024) $

// SIO4: Asynchronous Protocol Library: Sample Application: source file

#include "main.h"



//*****************************************************************************
static int _verify_model(const args_t* args, int* skip)
{
	int	errs;
	s32	z16c30;

	errs	= sio4_query(args->fd, -1, 1, SIO4_QUERY_MODEL_Z16C30, &z16c30);

	if (errs)
	{
		skip[0]	= 1;
	}
	else if (z16c30)
	{
		skip[0]	= 0;
	}
	else
	{
		skip[0]	= 1;
		gsc_label(NULL);
		printf("SKIP  (Not a Z16C30 model SIO4.)\n");
	}

	return(errs);
}



//*****************************************************************************
static int _initialize(const args_t* args, s32 baud, sio4_async_t* async)
{
	int					errs;
	sio4_async_init_t	init;
	const char*			msg		= NULL;
	int					mult;
	s32					osc;
	int					ret;

	gsc_label("Initialize");

	errs	= sio4_query(args->fd, -1, 0, SIO4_QUERY_OSC_PROGRAM, &osc);

	if (errs == 0)
	{
		init.tx_bit_rate	= baud;
		init.rx_bit_rate	= baud;
		init.osc_prog		= 20000000;

		if (args->pc)
		{
			init.tx_bit_rate	= 19200;
			init.rx_bit_rate	= 19200;
		}

		if (osc)
		{
			mult			= init.osc_prog / baud;
			init.osc_prog	= baud * mult;
		}

		ret		= sio4_async_init_data(args->fd, &init, async, &msg);
		errs	= ret ? 1 : 0;
	}

	if (errs == 0)
	{
		printf("PASS\n");
	}
	else
	{
		printf("FAIL <---");

		if (msg)
			printf("  (%s)", msg);

		printf("\n");
	}

	return(errs);
}



//*****************************************************************************
static int _config(const args_t* args, sio4_async_t* async)
{
	int			errs	= 0;
	const char*	msg		= NULL;
	int			ret;

	gsc_label("Configure");

	// The "async" structure has already been initialized. The USC substructure
	// has already beeen initialized for ASYNC operation at the specified baud
	// rate. There is little we need do here but change only what is needed.

	async->tx.io.timeout		= 1;	// In our case we'll write for X seconds then quit.
	async->cable.mode			= SIO4_ASYNC_CABLE_MODE_DTE;
	async->cable.protocol		= args->xcvr;
	async->cable.term.enable	= args->term_res;

	async->cable.legacy.txc		= SIO4_ASYNC_CABLE_LEGACY_TXC_UP;
	async->cable.legacy.txd_cts	= SIO4_ASYNC_CABLE_LEGACY_TXD_CTS_UP;
	async->cable.legacy.rxc		= SIO4_ASYNC_CABLE_LEGACY_RXC_DISABLE;
	async->cable.legacy.rxd_dcd	= SIO4_ASYNC_CABLE_LEGACY_RXD_DCD_DISABLE;
	async->usc.cts_legacy		= SIO4_ASYNC_USC_CTS_TRI;
	async->usc.dcd_legacy		= SIO4_ASYNC_USC_DCD_LEG_OUT_0;
	async->usc.tx.txc_legacy	= SIO4_ASYNC_USC_TX_TXC_LEG_OUT_CLK;
	async->usc.rx.rxc_legacy	= SIO4_ASYNC_USC_RX_RXC_LEG_IN;

	if (args->pc)
	{
		async->cable.enable		= SIO4_ASYNC_CABLE_ENABLE_YES;
		async->cable.mode		= SIO4_ASYNC_CABLE_MODE_DCE;
		async->cable.protocol	= SIO4_XCVR_PROTOCOL_RS232;
		async->tx.char_len		= SIO4_ASYNC_TX_CHAR_LEN_8;
		async->tx.encoding		= SIO4_ASYNC_TX_ENCODING_NRZ;
		async->tx.parity.enable	= SIO4_ASYNC_TX_PARITY_ENABLE_NO;
	}

	ret	= sio4_async_set(args->fd, async, &msg);

	if (ret)
	{
		errs++;
		printf("FAIL <---  (%s)\n", msg);
	}
	else if (args->pc)
	{
		printf("PASS  (PC: DCE, 19200 baud, 8 data bits, No Parity, NRZ)\n");
	}
	else
	{
		printf("PASS  (%ld baud)\n", (long) async->rx.bit_rate);
	}

	if (args->verbose)
	{
		ret	= sio4_async_show(args->fd, async, stdout, &msg);

		if (ret)
		{
			errs++;
			printf("FAIL <---  (%s)\n", msg);
		}
	}

	return(errs);
}



//*****************************************************************************
static void _delay_before(const args_t* args)
{
	if (args->delay_before)
	{
		gsc_label("Delay Before");
		printf("%d ms ...", (int) args->delay_before);
		fflush(stdout);
		os_sleep_ms(args->delay_before);
		printf("Done\n");
		fflush(stdout);
	}
}



//*****************************************************************************
static void _delay_after(const args_t* args)
{
	if (args->delay_after)
	{
		gsc_label("Delay After");
		printf("%d ms ...", (int) args->delay_after);
		fflush(stdout);
		os_sleep_ms(args->delay_after);
		printf("Done\n");
		fflush(stdout);
	}
}



//*****************************************************************************
static int _wait_til_done(const args_t* args)
{
	int	errs	= 0;
	s32	get;
	int	i;
	s32	last_1	= -1;
	s32	last_2	= -1;
	u32	tcsr;

	for (;;)	// A convenience loop.
	{
		if (args->wait_til_done == 0)
			break;

		gsc_label("Wait Til Done");

		// Wait til the external Tx FIFO is empty.

		for (i = 0; errs == 0; i++)
		{
			errs	= sio4_tx_fifo_fill_level(args->fd, -1, 0, -1, &get);

			if (get == 0)
				break;

			if ((get == last_1) && (last_1 == last_2))
				errs++;

			os_sleep_ms(3);
			last_2	= last_1;
			last_1	= get;
		}

		// Wait til the USC Tx FIFO is empty.

		for (i = 0; errs == 0; i++)
		{
			errs	= sio4_reg_read(args->fd, -1, 0, SIO4_USC_TCSR, &tcsr);

			if (tcsr & 0x01)	// "Tx Empty" status set.
				break;

			if (i >= 1000)
				errs++;

			os_sleep_ms(1);
		}

		// Wait til the USC transmitter is done.

		for (i = 0; errs == 0; i++)
		{
			errs	= sio4_reg_read(args->fd, -1, 0, SIO4_USC_TCSR, &tcsr);

			if (tcsr & 0x04)	// "All Done" status set.
				break;

			if (i >= 1000)
				errs++;

			os_sleep_ms(1);
		}

		break;
	}

	printf("%s\n", errs ? "FAIL <---" : "Done");
	fflush(stdout);
	return(errs);
}



//*****************************************************************************
static int _xfer_data(const args_t* args)
{
	int			calls		= 0;
	int			errs		= 0;
	time_t		end;
	int			i;
	char		msg[1024]	= "";
	time_t		now;
	int			send		= -1;
	int			sent;
	const char*	src			= "The quick brown fox jumped over the lazy dog.\r\n";
	const char*	text		= src;
	long		total		= 0;

	for (;;)	// convenience loop
	{
		gsc_label("Writing Data");
		printf(	"%d second%s\n",
				(int) args->seconds,
				(args->seconds == 1) ? "" : "s");

		gsc_label("Working");
		printf("%s", args->print ? "\n" : "... ");
		fflush(stdout);

		end	= time(NULL) + args->seconds;

		for (; errs == 0;)
		{
			if (send <= 0)
			{
				src		= text;
				send	= strlen(src);
			}

			calls++;
			sent	= sio4_async_write(args->fd, src, send);

			if (sent < 0)
			{
				errs++;
				sprintf(msg, "Device write error (%d).", sent);
				continue;
			}

			if (args->print)
			{
				for (i = 0; i < sent; i++)
					printf("%c", (char) src[i]);
			}

			send	-= sent;
			src		+= sent;
			total	+= sent;
			now		= time(NULL);

			if (now > end)
				break;
		}

		printf("%s", args->print ? "\n" : "Done\n");
		break;
	}

	gsc_label("Status");
	printf("%d error%s\n", errs, (errs == 1) ? "" : "s");

	if (msg[0])
	{
		gsc_label("Status");
		printf("%s\n", msg);
	}

	gsc_label("Total");
	printf("%ld Byte%s\n", total, (total == 1) ? "" : "s");

	gsc_label("Write Calls");
	printf("%d\n", calls);

	gsc_label("Bytes per Call");
	printf("%.3f Bytes\n", (float) total / (calls ? calls : 1));

	return(errs);
}



//*****************************************************************************
int perform_tests(const args_t* args, int* skip)
{
	sio4_async_t	async;
	s32				baud	= 19200;
	int				errs;

	for (;;)	// A convenience loop.
	{
		errs	= _verify_model(args, skip);

		if ((errs) || (skip[0]))
			break;

		errs	= _initialize(args, baud, &async);

		if ((errs == 0) && (skip[0] == 0))
			errs	= _config(args, &async);

		if ((errs == 0) && (skip[0] == 0))
		{
			_delay_before(args);
			errs	+= _xfer_data(args);
			errs	+= _wait_til_done(args);
			_delay_after(args);
		}

		break;
	}

	return(errs);
}


