// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/driver/read.c $
// $Rev: 51095 $
// $Date: 2022-06-01 16:13:16 -0500 (Wed, 01 Jun 2022) $

// SIO4: Device Driver: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_rx_avail_csr
*
*	Purpose:
*
*		See how much DMA data is available using the Control/Status Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*		bytes	The number of bytes desired.
*
*	Returned:
*
*		>= 0	This is the number of bytes to read.
*
******************************************************************************/

static long _rx_avail_csr(chan_data_t* chan, long bytes)
{
	dev_data_t*	dev		= chan->dev;
	u32			csr;
	long		qty;

	csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

	if (csr & D12)
		qty	= bytes;
	else
		qty	= 0;

	return(qty);
}



/******************************************************************************
*
*	Function:	_rx_avail_csr_fsr
*
*	Purpose:
*
*		See how much DMA data is available using the Control/Status Register
*		and the FIFO Size Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*		bytes	The number of bytes desired.
*
*	Returned:
*
*		>= 0	This is the number of bytes to read.
*
******************************************************************************/

static long _rx_avail_csr_fsr(chan_data_t* chan, long bytes)
{
	dev_data_t*	dev		= chan->dev;
	u32			csr;
	long		qty;

	csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

	if ((csr & D12) == 0)
		qty	= 0;
	else if (bytes > (long) chan->cache.fifo_size_rx)
		qty	= chan->cache.fifo_size_rx;
	else
		qty	= bytes;

	return(qty);
}



/******************************************************************************
*
*	Function:	_rx_avail_csr_fsr_rar
*
*	Purpose:
*
*		See how much data is available using the Control/Status Register, the
*		FIFO Size Register and the Rx Almost Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of bytes available.
*
******************************************************************************/

static long _rx_avail_csr_fsr_rar(chan_data_t* chan)
{
	dev_data_t*	dev		= chan->dev;
	u32			ae;
	u32			af;
	u32			csr;
	long		qty;
	u32			rar;

	for (;;)	// A convenience loop.
	{
		csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

		if ((csr & D15) == 0)
		{
			// The Rx FIFO is Full.
			qty	= chan->cache.fifo_size_rx;
			break;
		}

		if ((csr & D14) == 0)
		{
			// The Rx FIFO is Almost Full.
			rar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_rar_32);
			af	= rar >> 16;
			qty	= chan->cache.fifo_size_rx - af;
			break;
		}

		if (csr & D13)
		{
			// The Rx FIFO is more than Almost Empty.
			rar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_rar_32);
			ae	= rar & 0xFFFF;
			qty	= ae;
			break;
		}

		if (csr & D12)
		{
			// The Rx FIFO is above Empty.
			qty	= 1;
			break;
		}

		// The Rx FIFO is empty.
		qty	= 0;
		break;
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	_rx_avail_csr_rar
*
*	Purpose:
*
*		See how much data is available using the Control/Status Register and
*		the Rx Almost Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of bytes available.
*
******************************************************************************/

static long _rx_avail_csr_rar(chan_data_t* chan)
{
	dev_data_t*	dev		= chan->dev;
	u32			ae;
	u32			csr;
	long		qty;
	u32			rar;

	for (;;)	// A convenience loop.
	{
		csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

		if (csr & D13)
		{
			// The Rx FIFO is more than Almost Empty.
			rar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_rar_32);
			ae	= rar & 0xFFFF;
			qty	= ae;
			break;
		}

		if (csr & D12)
		{
			// The Rx FIFO is more than Empty.
			qty	= 1;
			break;
		}

		// The Rx FIFO is empty.
		qty	= 0;
		break;
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	_rx_avail_fcr
*
*	Purpose:
*
*		See how much data is available using the FIFO Count Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of bytes available.
*
******************************************************************************/

static long _rx_avail_fcr(chan_data_t* chan)
{
	dev_data_t*	dev		= chan->dev;
	u32			fcr;
	long		qty;

	fcr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_fcr_32);
	qty	= fcr >> 16;
	return(qty);
}



//*****************************************************************************
static void _dev_io_sw_init(chan_data_t* chan, dev_io_t* io)
{
	io->dma_threshold		= SIO4_IO_DMA_THRESHOLD_DEFAULT;
	io->io_mode				= SIO4_IO_MODE_DEFAULT;
	io->io_overrun_check	= SIO4_IO_OVERRUN_CHECK;
	io->io_underrun_check	= SIO4_IO_UNDERRUN_CHECK;
	io->pio_threshold		= SIO4_IO_PIO_THRESHOLD_DEFAULT;
	io->timeout_s			= SIO4_IO_TIMEOUT_DEFAULT;
}



//*****************************************************************************
static void _dev_io_close(chan_data_t* chan, dev_io_t* io)
{
	io->dma_channel	= NULL;
}



//*****************************************************************************
static void _dev_io_open(chan_data_t* chan, dev_io_t* io)
{
	_dev_io_sw_init(chan, io);
}



//*****************************************************************************
static int _dev_io_startup(chan_data_t* chan, dev_io_t* io)
{
	u32			bcr;
	u32			csr;
	dev_data_t*	dev		= chan->dev;
	long		ret		= 0;

	io->dma_threshold_last	= 0;

	if (((chan->cache.rx_fifo_overrun) && (io->io_overrun_check))	||
		((dev->cache.rx_fifo_underrun) && (io->io_underrun_check)))
	{
		csr	= os_reg_mem_rx_u32(chan->dev, chan->vaddr.gsc_csr_32);

		if ((chan->cache.rx_fifo_overrun) && (io->io_overrun_check) && (csr & D16))
			ret	= -EIO;

		if ((dev->cache.rx_fifo_underrun) && (io->io_underrun_check) && (csr & D18))
			ret	= -EIO;
	}

	if ((io->io_mode == GSC_IO_MODE_DMDMA) || (dev->cache.dmdma_scd))
	{
		bcr	= os_reg_mem_rx_u32(chan->dev, dev->vaddr.gsc_bcr_32);

		if (((bcr & D3) == 0) || ((bcr & D7) == 0))
		{
			bcr	|= D3 | D7;
			os_reg_mem_tx_u32(dev, dev->vaddr.gsc_bcr_32, bcr);
		}
	}

	return(ret);
}



//*****************************************************************************
static long _dev_pio_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	dev_data_t*	dev	= chan->dev;
	long		qty;

	if (dev->cache.reg_fcr)
		qty	= _rx_avail_fcr(chan);
	else if (dev->cache.reg_fsr)
		qty	= _rx_avail_csr_fsr_rar(chan);
	else
		qty	= _rx_avail_csr_rar(chan);

	if (qty > (long) bytes)
		qty	= bytes;

	return(qty);
}



//*****************************************************************************
static long _dev_pio_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	qty	= gsc_read_pio_work_8_bit(chan, io, mem, bytes, st_end);
	return(qty);
}



//*****************************************************************************
static long _dma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	dev_data_t*	dev	= chan->dev;
	long		qty;

	if (dev->cache.reg_fcr)
		qty	= _rx_avail_fcr(chan);
	else if (dev->cache.reg_fsr)
		qty	= _rx_avail_csr_fsr(chan, bytes);
	else
		qty	= _rx_avail_csr(chan, bytes);

	if (qty > (long) bytes)
	{
		qty	= (long) bytes;
	}
	else if (io->dma_threshold)
	{
		if (qty < io->dma_threshold)
		{
			if (qty != io->dma_threshold_last)
			{
				io->dma_threshold_last	= qty;
				qty	= 0;
			}
			else
			{
				io->dma_threshold_last	= 0;
			}
		}
	}

	if (qty > io->pio_threshold)
		qty	&= ~0x3L;	// Force it to a 32-bit boundary.

	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	long	qty;

	qty	= _dma_available(chan, io, bytes);
	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty		= 0;
	gsc_dma_setup_t	setup;

	if ((bytes < 8)			&&	// Don't force this option if either threshold is zero,
		(io->dma_threshold)	&&	// as this indicates an application is conducting tests.
		(io->pio_threshold))
	{
		qty	= gsc_read_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else if ((long) bytes < (long) io->pio_threshold)
	{
		qty	= gsc_read_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= chan;
		setup.dev		= chan->dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_READ;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_SIZE_8_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_BOARD_TO_HOST
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	if ((io->timeout_s == 0) || (io->frame.dev_dmdma_xfer))
		bytes	= _dma_available(chan, io, bytes);

	return(bytes);
}



//*****************************************************************************
static long _dev_dmdma_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty		= 0;
	gsc_dma_setup_t	setup;

	if ((bytes < 16) ||
		((io->pio_threshold > 0) && (bytes <= io->pio_threshold)))
	{
		qty	= gsc_read_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else
	{
		bytes	-= 8;
		bytes	&= ~0x7L;	// Force it to a 64-bit boundary.

		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= chan;
		setup.dev		= chan->dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_DMDMA_READ;

		setup.mode		= GSC_DMA_MODE_DM_DMA
						| GSC_DMA_MODE_SIZE_8_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_BOARD_TO_HOST
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	dev_read_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Read streaming I/O structure.
*
*	Arguments:
*
*		dev		The data for the device of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_read_create(chan_data_t* chan, dev_io_t* io)
{
	static const u32	_rx_dmdma_dir[4]	= { 0x0, 0x2, 0x1, 0x3 };

	int	ret;

	io->bytes_per_sample	= 1;
	io->dmdma_dir			= _rx_dmdma_dir[chan->index];
	io->io_reg_offset		= 0x18 + (chan->index * 16);
	io->io_reg_vaddr		= chan->vaddr.gsc_fdr_8;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available;
	io->dev_bmdma_available	= _dev_bmdma_available;
	io->dev_dmdma_available	= _dev_dmdma_available;
	io->dev_pio_xfer		= _dev_pio_xfer;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer;

	io->wait.abort			= SIO4_WAIT_IO_RX_ABORT;
	io->wait.done			= SIO4_WAIT_IO_RX_DONE;
	io->wait.error			= SIO4_WAIT_IO_RX_ERROR;
	io->wait.timeout		= SIO4_WAIT_IO_RX_TIMEOUT;

	ret	= gsc_io_create(chan, io, _64K);
	return(ret);
}


