// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/driver/write.c $
// $Rev: 51095 $
// $Date: 2022-06-01 16:13:16 -0500 (Wed, 01 Jun 2022) $

// SIO4: Device Driver: source file

#include "main.h"



/******************************************************************************
*
*	Function:	_tx_avail_csr
*
*	Purpose:
*
*		See how much DMA space is available using the Control/Status Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*		bytes	The number of bytes desired.
*
*	Returned:
*
*		>= 0	This is the number of bytes to write.
*
******************************************************************************/

static long _tx_avail_csr(chan_data_t* chan, long bytes)
{
	dev_data_t*	dev = chan->dev;
	u32			csr;
	long		qty;

	csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

	if (csr & D11)
		qty	= bytes;
	else
		qty	= 0;

	return(qty);
}



/******************************************************************************
*
*	Function:	_tx_avail_csr_fsr
*
*	Purpose:
*
*		See how much DMA space is available using the Control/Status Register
*		and the FIFO Size Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*		bytes	The number of bytes desired.
*
*	Returned:
*
*		>= 0	This is the number of bytes to write.
*
******************************************************************************/

static long _tx_avail_csr_fsr(chan_data_t* chan, long bytes)
{
	dev_data_t*	dev = chan->dev;
	u32			csr;
	long		qty;

	csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

	if ((csr & D11) == 0)
		qty	= 0;
	else if (bytes > (long) chan->cache.fifo_size_tx)
		qty	= chan->cache.fifo_size_tx;
	else
		qty	= bytes;

	return(qty);
}



/******************************************************************************
*
*	Function:	_tx_avail_csr_fsr_tar
*
*	Purpose:
*
*		See how much space is available using the Control/Status Register, the
*		FIFO Size Register and the Tx Almost Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of spaces available.
*
******************************************************************************/

static long _tx_avail_csr_fsr_tar(chan_data_t* chan)
{
	dev_data_t*	dev	= chan->dev;
	u32			ae;
	u32			af;
	u32			csr;
	long		qty;
	u32			tar;

	for (;;)	// A convenience loop.
	{
		csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

		if ((csr & D8) == 0)
		{
			// The Tx FIFO is Empty.
			qty	= chan->cache.fifo_size_tx;
			break;
		}

		if ((csr & D9) == 0)
		{
			// The Tx FIFO is Almost Empty.
			tar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_tar_32);
			ae	= tar & 0xFFFF;
			qty	= chan->cache.fifo_size_tx - ae;
			break;
		}

		if (csr & D10)
		{
			// The Tx FIFO is less than Almost Full.
			tar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_tar_32);
			af	= tar >> 16;
			qty	= af;
			break;
		}

		if (csr & D11)
		{
			// The Tx FIFO is less than Full.
			qty	= 1;
			break;
		}

		// The Tx FIFO is Full.
		qty	= 0;
		break;
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	_tx_avail_csr_tar
*
*	Purpose:
*
*		See how much space is available using the Control/Status Register and
*		the Tx Almost Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of bytes to write.
*
******************************************************************************/

static long _tx_avail_csr_tar(chan_data_t* chan)
{
	dev_data_t*	dev	= chan->dev;
	u32			af;
	u32			csr;
	long		qty;
	u32			tar;

	for (;;)	// A convenience loop.
	{
		csr	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_csr_32);

		if (csr & D10)
		{
			// The Tx FIFO is less than Almost Full.
			tar	= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_tar_32);
			af	= tar >> 16;
			qty	= af;
			break;
		}

		if (csr & D11)
		{
			// The Tx FIFO is less than Full.
			qty	= 1;
			break;
		}

		// The Tx FIFO is Full.
		qty	= 0;
		break;
	}

	return(qty);
}



/******************************************************************************
*
*	Function:	_tx_avail_fcr_fsr
*
*	Purpose:
*
*		See how much space is available using the FIFO Count Register and the
*		FIFO Size Register.
*
*	Arguments:
*
*		chan	The channel data structure.
*
*	Returned:
*
*		>= 0	This is the number of bytes to write.
*
******************************************************************************/

static long _tx_avail_fcr_fsr(chan_data_t* chan)
{
	dev_data_t*	dev	= chan->dev;
	u32			bytes;
	u32			fcr;
	long		qty;

	fcr		= os_reg_mem_rx_u32(dev, chan->vaddr.gsc_fcr_32);
	bytes	= fcr & 0xFFFF;
	qty		= chan->cache.fifo_size_tx - bytes;
	return(qty);
}



//*****************************************************************************
static void _dev_io_sw_init(chan_data_t* chan, dev_io_t* io)
{
	io->dma_threshold		= SIO4_IO_DMA_THRESHOLD_DEFAULT;
	io->io_mode				= SIO4_IO_MODE_DEFAULT;
	io->io_overrun_check	= SIO4_IO_OVERRUN_CHECK;
	io->io_underrun_check	= 0;	// not supported on Tx
	io->pio_threshold		= SIO4_IO_PIO_THRESHOLD_DEFAULT;
	io->timeout_s			= SIO4_IO_TIMEOUT_DEFAULT;
	io->frame.protocol		= -1;
}



//*****************************************************************************
static void _dev_io_close(chan_data_t* chan, dev_io_t* io)
{
	io->dma_channel	= NULL;
}



//*****************************************************************************
static void _dev_io_open(chan_data_t* chan, dev_io_t* io)
{
	_dev_io_sw_init(chan, io);
}



//*****************************************************************************
static int _dev_io_startup(chan_data_t* chan, dev_io_t* io)
{
	u32			bcr;
	u32			csr;
	dev_data_t*	dev	= chan->dev;
	long		ret	= 0;

	io->dma_threshold_last	= 0;

	if ((dev->cache.tx_fifo_overrun) && (io->io_overrun_check))
	{
		csr	= os_reg_mem_rx_u32(chan->dev, chan->vaddr.gsc_csr_32);

		if (csr & D17)
			ret	= -EIO;
	}

	if ((io->io_mode == GSC_IO_MODE_DMDMA) && (dev->cache.dmdma_scd))
	{
		// Disable Single Cycle operation
		bcr	= os_reg_mem_rx_u32(chan->dev, dev->vaddr.gsc_bcr_32);

		if (((bcr & D3) == 0) || ((bcr & D7) == 0))
		{
			bcr	|= D3 | D7;
			os_reg_mem_tx_u32(dev, dev->vaddr.gsc_bcr_32, bcr);
		}
	}

	return(ret);
}



//*****************************************************************************
long dev_pio_write_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	dev_data_t*	dev	= chan->dev;
	long		qty;

	if (dev->cache.reg_fsr == 0)
		qty	= _tx_avail_csr_tar(chan);
	else if (dev->cache.reg_fcr)
		qty	= _tx_avail_fcr_fsr(chan);
	else
		qty	= _tx_avail_csr_fsr_tar(chan);

	return(qty);
}



//*****************************************************************************
static long _dev_pio_available_sel(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	long	qty;

	if (io->frame.dev_pio_available)
		qty	= (io->frame.dev_pio_available)(chan, io, bytes);
	else
		qty	= dev_pio_write_available(chan, io, bytes);

	return(qty);
}



//*****************************************************************************
long dev_pio_write_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	qty	= gsc_write_pio_work_8_bit(chan, io, mem, bytes, st_end);
	return(qty);
}



//*****************************************************************************
static long _dev_pio_xfer_sel(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	if (io->frame.dev_pio_xfer)
		qty	= (io->frame.dev_pio_xfer)(chan, io, mem, bytes, st_end);
	else
		qty	= dev_pio_write_xfer(chan, io, mem, bytes, st_end);

	return(qty);
}



//*****************************************************************************
static long _dma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	dev_data_t*	dev	= chan->dev;
	long		qty;

	if (dev->cache.reg_fsr == 0)
		qty	= _tx_avail_csr(chan, bytes);
	else if (dev->cache.reg_fcr)
		qty	= _tx_avail_fcr_fsr(chan);
	else
		qty	= _tx_avail_csr_fsr(chan, bytes);

	if (qty > (long) bytes)
	{
		qty	= (long) bytes;
	}
	else if (io->dma_threshold)
	{
		if (qty < io->dma_threshold)
		{
			if (qty != io->dma_threshold_last)
			{
				io->dma_threshold_last	= qty;
				qty	= 0;
			}
			else
			{
				io->dma_threshold_last	= 0;
			}
		}
	}

	if (qty > io->pio_threshold)
		qty	&= ~0x3L;	// Force it to a 32-bit boundary.

	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	long	qty;

	qty	= _dma_available(chan, io, bytes);
	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_available_sel(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	long	qty;

	if (io->frame.dev_bmdma_available)
		qty	= (io->frame.dev_bmdma_available)(chan, io, bytes);
	else
		qty	= _dev_bmdma_available(chan, io, bytes);

	return(qty);
}



//*****************************************************************************
long dev_bmdma_tx_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty		= 0;
	gsc_dma_setup_t	setup;

	if ((bytes < 8)			&&	// Don't force this option if either threshold is zero,
		(io->dma_threshold)	&&	// as this indicates an application is conducting tests.
		(io->pio_threshold))
	{
		qty	= gsc_write_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else if ((long) bytes < (long) io->pio_threshold)
	{
		qty	= gsc_write_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else
	{
		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= chan;
		setup.dev		= chan->dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_BMDMA_WRITE;

		setup.mode		= GSC_DMA_MODE_BLOCK_DMA
						| GSC_DMA_MODE_SIZE_8_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_HOST_TO_BOARD
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_bmdma_xfer_sel(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	if (io->frame.dev_bmdma_xfer)
		qty	= (io->frame.dev_bmdma_xfer)(chan, io, mem, bytes, st_end);
	else
		qty	= dev_bmdma_tx_xfer(chan, io, mem, bytes, st_end);

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_available(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	if ((io->timeout_s == 0) || (io->frame.dev_dmdma_xfer))
		bytes	= _dma_available(chan, io, bytes);

	return(bytes);
}



//*****************************************************************************
static long _dev_dmdma_available_sel(chan_data_t* chan, dev_io_t* io, size_t bytes)
{
	long	qty;

	if (io->frame.dev_dmdma_available)
		qty	= (io->frame.dev_dmdma_available)(chan, io, bytes);
	else
		qty	= _dev_dmdma_available(chan, io, bytes);

	return(qty);
}



//*****************************************************************************
long dev_dmdma_tx_xfer(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long			qty		= 0;
	gsc_dma_setup_t	setup;

	if ((bytes < 16) ||
		((io->pio_threshold > 0) && (bytes <= io->pio_threshold)))
	{
		qty	= gsc_write_pio_work_8_bit(chan, io, mem, bytes, st_end);
	}
	else
	{
		bytes	-= 8;
		bytes	&= ~0x7L;	// Force it to a 64-bit boundary.

		memset(&setup, 0, sizeof(gsc_dma_setup_t));
		setup.alt		= chan;
		setup.dev		= chan->dev;
		setup.io		= io;
		setup.mem		= mem;
		setup.st_end	= st_end;
		setup.bytes		= bytes;
		setup.ability	= GSC_DMA_CAP_DMDMA_WRITE;

		setup.mode		= GSC_DMA_MODE_DM_DMA
						| GSC_DMA_MODE_SIZE_8_BITS
						| GSC_DMA_MODE_INPUT_ENABLE
						| GSC_DMA_MODE_BURSTING_LOCAL
						| GSC_DMA_MODE_INTERRUPT_WHEN_DONE
						| GSC_DMA_MODE_LOCAL_ADRESS_CONSTANT
						| GSC_DMA_MODE_PCI_INTERRUPT_ENABLE;

		setup.dpr		= GSC_DMA_DPR_HOST_TO_BOARD
						| GSC_DMA_DPR_END_OF_CHAIN
						| GSC_DMA_DPR_TERMINAL_COUNT_IRQ;

		qty	= gsc_dma_perform(&setup);
	}

	return(qty);
}



//*****************************************************************************
static long _dev_dmdma_xfer_sel(
	chan_data_t*	chan,
	dev_io_t*		io,
	const os_mem_t*	mem,
	size_t			bytes,
	os_time_tick_t	st_end)
{
	long	qty;

	if (io->frame.dev_dmdma_xfer)
		qty	= (io->frame.dev_dmdma_xfer)(chan, io, mem, bytes, st_end);
	else
		qty	= dev_dmdma_tx_xfer(chan, io, mem, bytes, st_end);

	return(qty);
}



/******************************************************************************
*
*	Function:	dev_write_create
*
*	Purpose:
*
*		Perform a one-tine setup of the Write streaming I/O structure.
*
*	Arguments:
*
*		dev		The data for the device of interest.
*
*		io		The I/O structure for this stream.
*
*	Returned:
*
*		None.
*
******************************************************************************/

int dev_write_create(chan_data_t* chan, dev_io_t* io)
{
	static const u32	_tx_dmdma_dir[4]	= { 0x4, 0x6, 0x5, 0x7 };

	int	ret;

	io->bytes_per_sample	= 1;
	io->dmdma_dir			= _tx_dmdma_dir[chan->index];
	io->io_reg_offset		= 0x18 + (chan->index * 16);
	io->io_reg_vaddr		= chan->vaddr.gsc_fdr_8;

	io->dev_io_sw_init		= _dev_io_sw_init;
	io->dev_io_close		= _dev_io_close;
	io->dev_io_open			= _dev_io_open;
	io->dev_io_startup		= _dev_io_startup;
	io->dev_pio_available	= _dev_pio_available_sel;
	io->dev_bmdma_available	= _dev_bmdma_available_sel;
	io->dev_dmdma_available	= _dev_dmdma_available_sel;
	io->dev_pio_xfer		= _dev_pio_xfer_sel;
	io->dev_bmdma_xfer		= _dev_bmdma_xfer_sel;
	io->dev_dmdma_xfer		= _dev_dmdma_xfer_sel;

	io->frame.dev_pio_available		= NULL;
	io->frame.dev_bmdma_available	= NULL;
	io->frame.dev_dmdma_available	= NULL;
	io->frame.dev_pio_xfer			= NULL;
	io->frame.dev_bmdma_xfer		= NULL;
	io->frame.dev_dmdma_xfer		= NULL;

	io->wait.abort			= SIO4_WAIT_IO_TX_ABORT;
	io->wait.done			= SIO4_WAIT_IO_TX_DONE;
	io->wait.error			= SIO4_WAIT_IO_TX_ERROR;
	io->wait.timeout		= SIO4_WAIT_IO_TX_TIMEOUT;

	ret	= gsc_io_create(chan, io, _64K);
	return(ret);
}


