// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/hdlc/api/rx_frame.c $
// $Rev: 52181 $
// $Date: 2022-12-13 12:22:03 -0600 (Tue, 13 Dec 2022) $

// SIO4: HDLC Protocol Library: source file

#include "main.h"



//*****************************************************************************
static int _args_validate(const sio4_hdlc_rx_frame_t* rx)
{
	int	ret	= -EINVAL;

	for (;;)	// A convenience loop.
	{
		if (rx == NULL)
			break;

		if (rx->status)
			break;

		if (rx->size)
			break;

		if (rx->last)
			break;

		if (rx->received)
			break;

		if (rx->buffer[0])
			break;

		ret	= 0;
		break;
	}

	return(ret);
}



//*****************************************************************************
static int _compute_read_size(hdlc_device_t* dev)
{
	int	need;
	int	ret		= 0;

	if (dev->rx.frame_size < 0)
	{
		// Frame size not yet known.

		if (dev->rx.on_hand > dev->rx.max_size)
		{
			// This is an error condition.
			ret				= -EOVERFLOW;
			dev->rx.line	= __LINE__;
		}
		else
		{
			// Request enough data to fill the buffer.
			dev->rx.get	= dev->rx.max_size - dev->rx.on_hand;
		}
	}
	else if (dev->rx.frame_size > 0)
	{
		// The frame size is known.
		need	= dev->rx.frame_size;

		if (dev->rx.on_hand >= need)
		{
			dev->rx.get	= 0;
		}
		else
		{
			dev->rx.get	= need - dev->rx.on_hand;
		}
	}
	else	// dev->rx.frame_size == 0
	{
		// There has been an issue.
		dev->rx.get	= 0;
	}

	return(ret);
}



//*****************************************************************************
// Copy the data from our buffer to the frame.
// Move what remains to the start of our buffer.
static void _finish_frame_copy(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	int	bytes;

	if (dev->rx.on_hand > 0)
	{
		bytes	= rx->size;

		if (bytes > sizeof(rx->buffer))
			bytes	= sizeof(rx->buffer);

		if (bytes > dev->rx.on_hand)
			bytes	= dev->rx.on_hand;

		memcpy(rx->buffer, dev->rx.buffer, bytes);

		if (dev->rx.on_hand > bytes)
		{
			memmove(dev->rx.buffer,
					dev->rx.buffer + bytes,
					dev->rx.on_hand - bytes);
		}

		dev->rx.on_hand	-= bytes;
		rx->received	= bytes;
	}
	else
	{
		rx->received	= 0;
		rx->buffer[0]	= 0;
	}
}



//*****************************************************************************
// Return a frame with status information only.
static void _finish_frame_empty(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	rx->status		= dev->rx.status;
	rx->size		= 0;
	rx->received	= 0;
	rx->last		= -1;
	rx->buffer[0]	= 0;
}



//*****************************************************************************
// We received too much data for the frame.
static void _finish_frame_big(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	rx->status	= dev->rx.status | SIO4_HDLC_RX_STAT_ERROR_BIG;
	rx->size	= dev->rx.frame_size;
	rx->last	= -1;
	_finish_frame_copy(dev, rx);
	rx->size	-= dev->rx.frame_extra;
}



//*****************************************************************************
static void _finish_frame_full(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	// A complete frame was received.
	rx->size	= dev->rx.frame_size;
	_finish_frame_copy(dev, rx);
	rx->rsb		= (((u16) rx->buffer[rx->size - 2]) << 0)
				| (((u16) rx->buffer[rx->size - 1]) << 8);
	rx->last	= GSC_FIELD_DECODE(rx->rsb, 11, 9);
	rx->status	= dev->rx.status | SIO4_HDLC_RX_STAT_FRAME_FULL | rx->rsb;
	rx->size	-= dev->rx.frame_extra;
}



//*****************************************************************************
// We received too little data for the frame.
static void _finish_frame_part(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	rx->status	= dev->rx.status | SIO4_HDLC_RX_STAT_ERROR_PART;
	rx->size	= dev->rx.frame_size;
	rx->last	= -1;
	_finish_frame_copy(dev, rx);
	rx->size	-= dev->rx.frame_extra;
}



//*****************************************************************************
// Finish the frame so it can be returned to the application.
static int _finish_frame(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	if (dev->rx.frame_size <= 0)
	{
		_finish_frame_empty(dev, rx);
	}
	else if (dev->rx.frame_size > 0xFFFF)
	{
		_finish_frame_big(dev, rx);
	}
	else if (dev->rx.on_hand >= dev->rx.frame_size)
	{
		_finish_frame_full(dev, rx);
	}
	else
	{
		_finish_frame_part(dev, rx);
	}

	return(0);
}



//*****************************************************************************
// Read data and append it to our buffer.
static int _read_data(hdlc_device_t* dev)
{
	u8*	dst;
	int	ret;

	for (;;)	// A convenience loop.
	{
		if (dev->rx.get < 0)
		{
			// There was a problem of some sort.
			dev->rx.got	= 0;
			ret			= dev->rx.get;	// negative
			break;
		}

		if (dev->rx.get == 0)
		{
			// There wasn't a problem, but there just isn't any data.
			dev->rx.got	= 0;
			ret			= 0;	// no data
			break;
		}

		dst	= dev->rx.buffer + dev->rx.on_hand;
		ret	= sio4_read(dev->fd, dst, dev->rx.get);

		if (ret < 0)
		{
			// There was a read error.
			dev->rx.got	= 0;
			break;
		}

		// No problems.
		dev->rx.on_hand	+= ret;
		dev->rx.got		= ret;
		break;
	}

	return(ret);
}



//*****************************************************************************
// Retrieve an RCC software FIFO entry, if available.
static int _sw_rcc_fifo_check(hdlc_device_t* dev)
{
	s32	arg;
	s32	enable;
	int	extra;
	u32	rccr;
	int	ret;
	int	tmp;
	s32	type;

	for (;;)	// A convenience loop.
	{
		// See if any records are available.
		arg	= -1;
		ret	= sio4_ioctl(dev->fd, SIO4_IOCTL_RCC_SW_FIFO_READ, &arg);

		if (ret < 0)
			break;

		if (arg == -1)
		{
			ret	= -ENOTSUP;
			break;
		}

		if (arg <= 0)
			break;

		// Retrieve the first available record.
		arg	= 1;
		ret	= sio4_ioctl(dev->fd, SIO4_IOCTL_RCC_SW_FIFO_READ, &arg);

		if (ret < 0)
			break;

		if (arg == -1)
		{
			ret	= -ENOTSUP;
			break;
		}

		if (arg & SIO4_RCC_SW_FIFO_READ_RX_FRAME)
		{
			enable	= -1;
			ret		= sio4_ioctl(dev->fd, SIO4_IOCTL_USC_RX_CRC_ENABLE, &enable);
			type	= -1;
			tmp		= sio4_ioctl(dev->fd, SIO4_IOCTL_USC_RX_CRC_TYPE, &type);
			ret		= ret ? ret : tmp;

			if (enable == SIO4_USC_CRC_ENABLE_YES)
				extra	= (type == SIO4_USC_CRC_TYPE_32) ? 4 : 2;
			else
				extra	= 0;

			extra	+= 2;	// Receive Status Block

			rccr				= arg & 0xFFFF;
			dev->rx.frame_size	= ((~rccr) & 0xFFFF) + 2;	// +2 for RSB
			dev->rx.frame_extra	= extra;
			dev->rx.status		= arg & 0xFFFF0000;
		}
		else
		{
			dev->rx.frame_size	= 0;
			dev->rx.frame_extra	= 0;
			dev->rx.status		= arg & 0xFFFF0000;
		}

		break;
	}

	return(ret);
}



//*****************************************************************************
static int _perform_operation(hdlc_device_t* dev, sio4_hdlc_rx_frame_t* rx)
{
	s32				arg;
	os_time_us_t	now;
	int				ret;
	os_time_us_t	stop;
	s32				timeout;
	int				tmp;

	ret	= os_sem_lock(&dev->rx.sem);

	if (ret == 0)
	{
		timeout	= -1;
		ret		= sio4_ioctl(dev->fd, SIO4_IOCTL_RX_IO_TIMEOUT, &timeout);

		if (ret == 0)
		{
			// We set the timeout to zero so the driver will give us what ever data
			// is available without blocking. Otherwise we won't be able to detect
			// the end of a frame in a timely manner.
			arg	= 0;
			ret	= sio4_ioctl(dev->fd, SIO4_IOCTL_RX_IO_TIMEOUT, &arg);
		}

		os_time_get_us(&stop);
		stop.tv_sec			+= timeout;	// the timeout may be zero
		dev->rx.frame_size	= -1;
		dev->rx.status		= 0;

		for (; ret >= 0;)
		{
			ret	= _compute_read_size(dev);

			if (ret < 0)
				break;

			if (dev->rx.get <= 0)
			{
				ret	= _finish_frame(dev, rx);
				break;
			}

			ret	= _read_data(dev);

			if (ret < 0)
				break;

			if (dev->rx.frame_size < 0)
			{
				ret	= _sw_rcc_fifo_check(dev);

				if (ret < 0)
					break;
			}

			if (dev->rx.frame_size == 0)
			{
				tmp	= _finish_frame(dev, rx);
				ret	= ret ? ret : tmp;
				break;
			}

			if (timeout <= 0)
			{
				if (dev->rx.got == 0)
				{
					// Quit as soon as no data is available from the device.
					tmp	= _finish_frame(dev, rx);
					ret	= ret ? ret : tmp;
					break;
				}

				continue;
			}

			// The timeout period is positive.
			os_time_get_us(&now);

			if ((now.tv_sec > stop.tv_sec) ||
				((now.tv_sec == stop.tv_sec) && (now.tv_usec >= stop.tv_usec)))
			{
				dev->rx.status |= SIO4_HDLC_RX_STAT_TIMEOUT;
				tmp	= _finish_frame(dev, rx);
				ret	= ret ? ret : tmp;
				break;
			}

			// Wait a moment to receive more data.
			os_sleep_ms(1);
		}

		// Restore the original timeout period.
		sio4_ioctl(dev->fd, SIO4_IOCTL_RX_IO_TIMEOUT, &timeout);

		os_sem_unlock(&dev->rx.sem);
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	sio4_hdlc_rx_frame
*
*	Purpose:
*
*		Read in a frame of data.
*
*	Arguments:
*
*		fd		The file descriptor for the SIO4 channel to access.
*
*		rx		The structure where the received frame is recorded.
*
*	Returned:
*
*		0		All went well.
*		-errno	There was a problem.
*
******************************************************************************/

int sio4_hdlc_rx_frame(int fd, sio4_hdlc_rx_frame_t* rx)
{
	hdlc_device_t*	dev	= NULL;
	int				ret;

	for (;;)	// A convenience loop.
	{
		// Verify Initialization ==========================
		ret	= hdlc_lib_init_verify(NULL);

		if (ret)
			break;

		// Access Device ==================================
		ret	= hdlc_lib_fd_inc(fd, &dev);

		if (ret)
			break;

		// Validate Arguments =============================
		ret	= _args_validate(rx);

		// Perform Operation ==============================

		if (ret)
		{
		}
		else if (dev->z16c30)
		{
			ret	= _perform_operation(dev, rx);
		}
		else
		{
			ret	= -EPROTONOSUPPORT;
		}

		// Release Device =================================
		hdlc_lib_fd_dec(dev);
		break;
	}

	return(ret);
}


