// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/SIO4%20and%20SIO8/SIO4_Linux_2.x.x_GSC_DN/hdlc/api/usc_cfg_rxc_ctr_brg_dpll_rxclk.c $
// $Rev: 50131 $
// $Date: 2022-02-04 09:04:14 -0600 (Fri, 04 Feb 2022) $

// SIO4: HDLC Protocol Library: source file

#include "main.h"

// NOTES: This module derives a USC configuration designed to produce the
// requested Rx bit rate via the USC RxC input pin and the DPLL. The process
// will attempt to use the minimum resources while reusing resources already
// in use. No resource settings already in use will be altered. This module
// evaluates all USC hardware options and all suitable configuration settings.



// macros *********************************************************************

#define	ABS(v)			(((v) < 0) ? (-(v)) : (v))

#define	CURRENT			0x10000	// current->usc.xxx device is used
#define	DESIRED			0x01000	// desired->usc.xxx device is used
#define	CLOCK_SOURCE	0x00100	// Both devices use the same clock source
#define	RATE			0x00010	// Both devices have the same rate and rate source.
#define	REUSE			0x00001	// reuse is expected

#define	_FLOAT_MAX_		(FLT_MAX / 2)



// data types *****************************************************************

typedef struct
{
	s32		enable;		// hdlc->usc.brg0.enable	yes, no, ignore
	s32		clk_src;	// hdlc->usc.brg0.clk_src	RxC, CTR0, CTR1, ignore
} brg0_t;

typedef struct
{
	s32		enable;		// hdlc->usc.brg1.enable	yes, no, ignore
	s32		clk_src;	// hdlc->usc.brg1.clk_src	RxC, CTR0, CTR1, ignore
} brg1_t;

typedef struct
{
	s32		clk_src;	// hdlc->usc.ctr0.clk_src	RxC, ignore
	s32		rate;		// hdlc->usc.ctr0.rate		32, 16, 8, 4, ignore
	s32		divide;		//							32, 16, 8, 4, ignore
}  ctr0_t;

typedef struct
{
	s32		clk_src;	// hdlc->usc.ctr1.clk_src	RxC, ignore
	s32		rate_src;	// hdlc->usc.ctr1.rate_src	CTR0, DPLL, ignore
}  ctr1_t;

typedef struct
{
	s32		clk_src;	// hdlc->usc.dpll.clk_src	RxC, BRG0, BRG1, ignore
	s32		rate;		// hdlc->usc.dpll.rate		32, 16, 8, ignore
	s32		divide;		//							32, 16, 8, ignore
}  dpll_t;

typedef struct
{
	const ctr0_t*	ctr0;
	const ctr1_t*	ctr1;
	const brg0_t*	brg0;
	const brg1_t*	brg1;
	const dpll_t*	dpll;
}  option_t;



// variables ******************************************************************

//										enable								clk_src
static const brg0_t	_brg0_ignore	= { -1,									-1									};
static const brg0_t	_brg0_ctr0		= { SIO4_HDLC_USC_BRG0_ENABLE_YES,		SIO4_HDLC_USC_BRG0_CLK_SRC_CTR0		};
static const brg0_t	_brg0_ctr1		= { SIO4_HDLC_USC_BRG0_ENABLE_YES,		SIO4_HDLC_USC_BRG0_CLK_SRC_CTR1		};
static const brg0_t	_brg0_rxc		= { SIO4_HDLC_USC_BRG0_ENABLE_YES,		SIO4_HDLC_USC_BRG0_CLK_SRC_RXC_PIN	};

//										enable								clk_src
static const brg1_t	_brg1_ignore	= { -1,									-1									};
static const brg1_t	_brg1_ctr0		= { SIO4_HDLC_USC_BRG1_ENABLE_YES,		SIO4_HDLC_USC_BRG1_CLK_SRC_CTR0		};
static const brg1_t	_brg1_ctr1		= { SIO4_HDLC_USC_BRG1_ENABLE_YES,		SIO4_HDLC_USC_BRG1_CLK_SRC_CTR1		};
static const brg1_t	_brg1_rxc		= { SIO4_HDLC_USC_BRG1_ENABLE_YES,		SIO4_HDLC_USC_BRG1_CLK_SRC_RXC_PIN	};

//										clk_src								rate							divide
static const ctr0_t	_ctr0_ignore	= {	-1,									-1,								-1		};
static const ctr0_t	_ctr0_div_4		= {	-1,									SIO4_HDLC_USC_CTR0_RATE_4X,		4		};
static const ctr0_t	_ctr0_div_8		= {	-1,									SIO4_HDLC_USC_CTR0_RATE_8X,		8		};
static const ctr0_t	_ctr0_div_16	= {	-1,									SIO4_HDLC_USC_CTR0_RATE_16X,	16		};
static const ctr0_t	_ctr0_div_32	= {	-1,									SIO4_HDLC_USC_CTR0_RATE_32X,	32		};
static const ctr0_t	_ctr0_rxc_4		= { SIO4_HDLC_USC_CTR0_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR0_RATE_4X,		4		};
static const ctr0_t	_ctr0_rxc_8		= { SIO4_HDLC_USC_CTR0_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR0_RATE_8X,		8		};
static const ctr0_t	_ctr0_rxc_16	= { SIO4_HDLC_USC_CTR0_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR0_RATE_16X,	16		};
static const ctr0_t	_ctr0_rxc_32	= { SIO4_HDLC_USC_CTR0_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR0_RATE_32X,	32		};

//										clk_src								rate_src
static const ctr1_t	_ctr1_ignore	= { -1,									-1										};
static const ctr1_t	_ctr1_rxc_ctr0	= { SIO4_HDLC_USC_CTR1_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR1_RATE_SRC_CTR0		};
static const ctr1_t	_ctr1_rxc_dpll	= { SIO4_HDLC_USC_CTR1_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_CTR1_RATE_SRC_DPLL		};

//										clk_src								rate							divide
static const dpll_t	_dpll_brg0_8	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG0,	SIO4_HDLC_USC_DPLL_RATE_8X,		8	};
static const dpll_t	_dpll_brg0_16	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG0,	SIO4_HDLC_USC_DPLL_RATE_16X,	16	};
static const dpll_t	_dpll_brg0_32	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG0,	SIO4_HDLC_USC_DPLL_RATE_32X,	32	};

static const dpll_t	_dpll_brg1_8	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG1,	SIO4_HDLC_USC_DPLL_RATE_8X,		8	};
static const dpll_t	_dpll_brg1_16	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG1,	SIO4_HDLC_USC_DPLL_RATE_16X,	16	};
static const dpll_t	_dpll_brg1_32	= { SIO4_HDLC_USC_DPLL_CLK_SRC_BRG1,	SIO4_HDLC_USC_DPLL_RATE_32X,	32	};

static const dpll_t	_dpll_rxc_8		= { SIO4_HDLC_USC_DPLL_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_DPLL_RATE_8X,		8	};
static const dpll_t	_dpll_rxc_16	= { SIO4_HDLC_USC_DPLL_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_DPLL_RATE_16X,	16	};
static const dpll_t	_dpll_rxc_32	= { SIO4_HDLC_USC_DPLL_CLK_SRC_RXC_PIN,	SIO4_HDLC_USC_DPLL_RATE_32X,	32	};

static const option_t _options[]	=
{
	// List Order: primarily fewer resources first, then alphabetical/numericsl

	// RxC -> DPLL -> RxClk								(See Notes: sheet A)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ignore,	&_dpll_rxc_8	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ignore,	&_dpll_rxc_16	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ignore,	&_dpll_rxc_32	},

	// RxC -> BRG0 -> DPLL -> RxClk						(See Notes: sheet B)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_rxc,		&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_rxc,		&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_rxc,		&_brg1_ignore,	&_dpll_brg0_32	},

	// RxC -> BRG1 -> DPLL -> RxClk						(See Notes: sheet C)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_rxc,		&_dpll_brg1_8	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_rxc,		&_dpll_brg1_16	},
	{ &_ctr0_ignore,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_rxc,		&_dpll_brg1_32	},

	// RxC -> CTR1(DPLL) -> BRG0 -> DPLL -> RxClk		(See Notes: sheet D)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_32	},

	// RxC -> CTR1(DPLL) -> BRG1 -> DPLL -> RxClk		(See Notes: sheet E)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_8	},
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_16	},
	{ &_ctr0_ignore,	&_ctr1_rxc_dpll,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_32	},

	// RxC -> CTR0 -> BRG0 -> DPLL -> RxClk				(See Notes: sheet F)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ctr0,	&_brg1_ignore,	&_dpll_brg0_32	},

	// RxC -> CTR0 -> BRG1 -> DPLL -> RxClk				(See Notes: sheet G)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_8	},
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_16	},
	{ &_ctr0_rxc_4,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_32	},

	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_8	},
	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_16	},
	{ &_ctr0_rxc_8,		&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_32	},

	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_8	},
	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_16	},
	{ &_ctr0_rxc_16,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_32	},

	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_8	},
	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_16	},
	{ &_ctr0_rxc_32,	&_ctr1_ignore,		&_brg0_ignore,	&_brg1_ctr0,	&_dpll_brg1_32	},

	// RxC -> CTR1(CTR0) -> BRG0 -> DPLL -> RxClk		(See Notes: sheet H)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_32	},

	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_8	},
	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_16	},
	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ctr1,	&_brg1_ignore,	&_dpll_brg0_32	},

	// RxC -> CTR1(CTR0) -> BRG1 -> DPLL -> RxClk		(See Notes: sheet I)
	// ctr0				ctr1				brg0			brg1			dpll
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_8	},
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_16	},
	{ &_ctr0_div_4,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_32	},

	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_8	},
	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_16	},
	{ &_ctr0_div_8,		&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_32	},

	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_8	},
	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_16	},
	{ &_ctr0_div_16,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_32	},

	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_8	},
	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_16	},
	{ &_ctr0_div_32,	&_ctr1_rxc_ctr0,	&_brg0_ignore,	&_brg1_ctr1,	&_dpll_brg1_32	}
};



//*****************************************************************************
static int _eval_state_5(int state)
{
	int	use;

	switch (state)
	{
		default:							// current	desired	dev		rate	reuse
		case 0x00000:	use	= 1;	break;	// unused	unused	!equal	!equal	no
		case 0x00001:	use	= 1;	break;	// unused	unused	!equal	!equal	yes
		case 0x00010:	use	= 1;	break;	// unused	unused	!equal	equal	no
		case 0x00011:	use	= 1;	break;	// unused	unused	!equal	equal	yes
		case 0x00100:	use	= 1;	break;	// unused	unused	equal	!equal	no
		case 0x00101:	use	= 1;	break;	// unused	unused	equal	!equal	yes
		case 0x00110:	use	= 1;	break;	// unused	unused	equal	equal	no
		case 0x00111:	use	= 1;	break;	// unused	unused	equal	equal	yes
		case 0x01000:	use	= 1;	break;	// unused	used	!equal	!equal	no
		case 0x01001:	use	= 0;	break;	// unused	used	!equal	!equal	yes
		case 0x01010:	use	= 1;	break;	// unused	used	!equal	equal	no
		case 0x01011:	use	= 0;	break;	// unused	used	!equal	equal	yes
		case 0x01100:	use	= 1;	break;	// unused	used	equal	!equal	no
		case 0x01101:	use	= 0;	break;	// unused	used	equal	!equal	yes
		case 0x01110:	use	= 1;	break;	// unused	used	equal	equal	no
		case 0x01111:	use	= 0;	break;	// unused	used	equal	equal	yes
		case 0x10000:	use	= 1;	break;	// used		unused	!equal	!equal	no
		case 0x10001:	use	= 0;	break;	// used		unused	!equal	!equal	yes
		case 0x10010:	use	= 1;	break;	// used		unused	!equal	equal	no
		case 0x10011:	use	= 0;	break;	// used		unused	!equal	equal	yes
		case 0x10100:	use	= 1;	break;	// used		unused	equal	!equal	no
		case 0x10101:	use	= 0;	break;	// used		unused	equal	!equal	yes
		case 0x10110:	use	= 1;	break;	// used		unused	equal	equal	no
		case 0x10111:	use	= 0;	break;	// used		unused	equal	equal	yes
		case 0x11000:	use	= 0;	break;	// used		used	!equal	!equal	no
		case 0x11001:	use	= 0;	break;	// used		used	!equal	!equal	yes
		case 0x11010:	use	= 0;	break;	// used		used	!equal	equal	no
		case 0x11011:	use	= 0;	break;	// used		used	!equal	equal	yes
		case 0x11100:	use	= 0;	break;	// used		used	equal	!equal	no
		case 0x11101:	use	= 0;	break;	// used		used	equal	!equal	yes
		case 0x11110:	use	= 1;	break;	// used		used	equal	equal	no
		case 0x11111:	use	= 1;	break;	// used		used	equal	equal	yes
	}

	return(use);
}



//*****************************************************************************
static int _brg0_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	int	state	= 0;
	int	use;		// Is this configuration option usable?

	if (option->brg0->clk_src < 0)
	{
		// This configuration is usable.
		use	= 1;
	}
	else
	{
		state	|= (current->usc.brg0.enable != SIO4_HDLC_USC_BRG0_ENABLE_NO)	? CURRENT		: 0;
		state	|= (desired->usc.brg0.enable != SIO4_HDLC_USC_BRG0_ENABLE_NO)	? DESIRED		: 0;
		state	|= (current->usc.brg0.clk_src == desired->usc.brg0.clk_src)		? CLOCK_SOURCE	: 0;
		state	|= (current->usc.brg0.divider == desired->usc.brg0.divider)		? RATE			: 0;
		state	|= reuse														? REUSE			: 0;
		use		= _eval_state_5(state);
	}

	return(use);
}



//*****************************************************************************
static int _brg1_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	int	state	= 0;
	int	use;		// Is this configuration option usable?

	if (option->brg1->clk_src < 0)
	{
		// This configuration is usable.
		use	= 1;
	}
	else
	{
		state	|= (current->usc.brg1.enable != SIO4_HDLC_USC_BRG1_ENABLE_NO)	? CURRENT		: 0;
		state	|= (desired->usc.brg1.enable != SIO4_HDLC_USC_BRG1_ENABLE_NO)	? DESIRED		: 0;
		state	|= (current->usc.brg1.clk_src == desired->usc.brg1.clk_src)		? CLOCK_SOURCE	: 0;
		state	|= (current->usc.brg1.divider == desired->usc.brg1.divider)		? RATE			: 0;
		state	|= reuse														? REUSE			: 0;
		use		= _eval_state_5(state);
	}

	return(use);
}



//*****************************************************************************
static int _ctr0_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	int	state	= 0;
	int	use;		// Is this configuration option usable?

	if (option->ctr0->clk_src < 0)
	{
		// This configuration is usable.
		use	= 1;
	}
	else
	{
		state	|= (current->usc.ctr0.clk_src != SIO4_HDLC_USC_CTR0_CLK_SRC_DISABLE)	? CURRENT		: 0;
		state	|= (desired->usc.ctr0.clk_src != SIO4_HDLC_USC_CTR0_CLK_SRC_DISABLE)	? DESIRED		: 0;
		state	|= (current->usc.ctr0.clk_src == desired->usc.ctr0.clk_src)				? CLOCK_SOURCE	: 0;
		state	|= (current->usc.ctr0.rate == desired->usc.ctr0.rate)					? RATE			: 0;
		state	|= reuse																? REUSE			: 0;
		use		= _eval_state_5(state);
	}

	return(use);
}



//*****************************************************************************
static int _ctr1_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	int	state	= 0;
	int	use;		// Is this configuration option usable?

	if (option->ctr1->clk_src < 0)
	{
		// This configuration is usable.
		use	= 1;
	}
	else
	{
		if (current->usc.ctr1.rate_src != desired->usc.ctr1.rate_src)
		{
		}
		else if (	(current->usc.ctr1.rate_src == SIO4_HDLC_USC_CTR1_RATE_SRC_CTR0)	&&
					(current->usc.ctr0.rate == desired->usc.ctr0.rate))
		{
			state	|= RATE;
		}
		else if (	(current->usc.ctr1.rate_src == SIO4_HDLC_USC_CTR1_RATE_SRC_DPLL)	&&
					(current->usc.dpll.rate == desired->usc.dpll.rate))
		{
			state	|= RATE;
		}

		state	|= (current->usc.ctr1.clk_src != SIO4_HDLC_USC_CTR1_CLK_SRC_DISABLE)	? CURRENT		: 0;
		state	|= (desired->usc.ctr1.clk_src != SIO4_HDLC_USC_CTR1_CLK_SRC_DISABLE)	? DESIRED		: 0;
		state	|= (current->usc.ctr1.clk_src == desired->usc.ctr1.clk_src)				? CLOCK_SOURCE	: 0;
		state	|= reuse																? REUSE			: 0;
		use		= _eval_state_5(state);
	}

	return(use);
}



//*****************************************************************************
static int _dpll_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	// The DPLL is used in all configurations.
	// We can change all settings as needed.
	// All DPLL options are thus usable.
	return(1);
}



//*****************************************************************************
static int _usc_check_usage(const sio4_hdlc_t* current, const sio4_hdlc_t* desired, const option_t* option, int reuse)
{
	int	brg0;		// Is this configuration option usable?
	int	brg1;		// Is this configuration option usable?
	int	ctr0;		// Is this configuration option usable?
	int	ctr1;		// Is this configuration option usable?
	int	dpll;		// Is this configuration option usable?
	int	use;		// Is this configuration option usable?

	brg0	= _brg0_check_usage(current, desired, option, reuse);
	brg1	= _brg1_check_usage(current, desired, option, reuse);
	ctr0	= _ctr0_check_usage(current, desired, option, reuse);
	ctr1	= _ctr1_check_usage(current, desired, option, reuse);
	dpll	= _dpll_check_usage(current, desired, option, reuse);
	use		= brg0 && brg1 && ctr0 && ctr1 && dpll;
	return(use);
}



//*****************************************************************************
static void _apply_option(sio4_hdlc_t* hdlc, const option_t* option)
{
	// BRG0

	if (option->brg0->enable >= 0)
		hdlc->usc.brg0.enable	= option->brg0->enable;

	if (option->brg0->clk_src >= 0)
		hdlc->usc.brg0.clk_src	= option->brg0->clk_src;

	// BRG1

	if (option->brg1->enable >= 0)
		hdlc->usc.brg1.enable	= option->brg1->enable;

	if (option->brg1->clk_src >= 0)
		hdlc->usc.brg1.clk_src	= option->brg1->clk_src;

	// CTR0

	if (option->ctr0->clk_src >= 0)
		hdlc->usc.ctr0.clk_src	= option->ctr0->clk_src;

	if (option->ctr0->rate >= 0)
		hdlc->usc.ctr0.rate		= option->ctr0->rate;

	// CTR1

	if (option->ctr1->clk_src >= 0)
		hdlc->usc.ctr1.clk_src	= option->ctr1->clk_src;

	if (option->ctr1->rate_src >= 0)
		hdlc->usc.ctr1.rate_src	= option->ctr1->rate_src;

	// DPLL

	if (option->dpll->clk_src >= 0)
		hdlc->usc.dpll.clk_src	= option->dpll->clk_src;

	if (option->dpll->rate >= 0)
		hdlc->usc.dpll.rate		= option->dpll->rate;
}



//*****************************************************************************
static void _compute_div_ctr0(const option_t* option, int* ctr)
{
	ctr[0]	= option->ctr0->divide;
}



//*****************************************************************************
static void _compute_div_ctr1(const option_t* option, int* ctr)
{
	switch (option->ctr1->rate_src)
	{
		default:

			break;

		case SIO4_HDLC_USC_CTR1_RATE_SRC_CTR0:

			ctr[0]	= option->ctr0->divide;
			break;

		case SIO4_HDLC_USC_CTR1_RATE_SRC_DPLL:

			ctr[0]	= option->dpll->divide;
			break;
	}
}



//*****************************************************************************
static void _compute_div_brg0(const option_t* option, int* brg, int* ctr0, int* ctr1)
{
	brg[0]	= 1;

	switch (option->brg0->clk_src)
	{
		default:

			break;

		case SIO4_HDLC_USC_BRG0_CLK_SRC_CTR0:

			_compute_div_ctr0(option, ctr0);
			break;

		case SIO4_HDLC_USC_BRG0_CLK_SRC_CTR1:

			_compute_div_ctr1(option, ctr1);
			break;
	}
}



//*****************************************************************************
static void _compute_div_brg1(const option_t* option, int* brg, int* ctr0, int* ctr1)
{
	brg[0]	= 1;

	switch (option->brg1->clk_src)
	{
		default:

			break;

		case SIO4_HDLC_USC_BRG1_CLK_SRC_CTR0:

			_compute_div_ctr0(option, ctr0);
			break;

		case SIO4_HDLC_USC_BRG1_CLK_SRC_CTR1:

			_compute_div_ctr1(option, ctr1);
			break;
	}
}



//*****************************************************************************
static void _compute_div_dpll(const option_t* option, int* dpll, int* brg0, int* brg1, int* ctr0, int* ctr1)
{
	dpll[0]	= option->dpll->divide;

	switch (option->dpll->clk_src)
	{
		default:

			break;

		case SIO4_HDLC_USC_DPLL_CLK_SRC_BRG0:

			_compute_div_brg0(option, brg0, ctr0, ctr1);
			break;

		case SIO4_HDLC_USC_DPLL_CLK_SRC_BRG1:

			_compute_div_brg1(option, brg1, ctr0, ctr1);
			break;
	}
}



//*****************************************************************************
static void _eval_brg_range(s32 bit_rate, const sio4_hdlc_t* hdlc, const option_t* option, int ctr0, int ctr1, int dpll, int* brg_min, int* brg_max)
{
	int	divide;

	if ((option->brg0->clk_src >= 0) &&
		(hdlc->usc.brg0.enable == SIO4_HDLC_USC_BRG0_ENABLE_YES))
	{
		brg_min[0]	= hdlc->usc.brg0.divider;
		brg_max[0]	= brg_min[0];
	}
	else
	if ((option->brg1->clk_src >= 0) &&
		(hdlc->usc.brg1.enable == SIO4_HDLC_USC_BRG1_ENABLE_YES))
	{
		brg_min[0]	= hdlc->usc.brg1.divider;
		brg_max[0]	= brg_min[0];
	}
	else
	{
		divide		= (int) ((float) hdlc->osc.prog / ctr0 /  ctr1 / dpll / bit_rate - 1);
		brg_min[0]	= divide - 3;
		brg_max[0]	= divide + 1;

		if (brg_min[0] < 0)
			brg_min[0]	= 0;
		else if (brg_min[0] > 0xFFFF)
			brg_min[0]	= 0xFFFF;

		if (brg_max[0] < 0)
			brg_max[0]	= 0;
		else if (brg_max[0] > 0xFFFF)
			brg_max[0]	= 0xFFFF;
	}
}



//*****************************************************************************
static float _compute_rx_rate(s32 bit_rate, const sio4_hdlc_t* ref, sio4_hdlc_t* hdlc, const option_t* option)
{
	int		brg			= 0;
	int		brg0		= 0;
	int		brg1		= 0;
	int		brg_max;
	int		brg_min;
	int		ctr0		= 0;
	int		ctr1		= 0;
	float	diff;
	float	diff_best	= _FLOAT_MAX_;
	int		dpll;
	float	rate		= 0;
	float	tmp;

	_compute_div_dpll(option, &dpll, &brg0, &brg1, &ctr0, &ctr1);

	if (ctr0 <= 0)
		ctr0	= 1;

	if (ctr1 <= 0)
		ctr1	= 1;

	if ((brg0 <= 0) && (brg1 <= 0))
	{
		rate	= (float) hdlc->osc.prog / ctr0 /  ctr1 / dpll;
	}
	else
	{
		_eval_brg_range(bit_rate, ref, option, ctr0, ctr1, dpll, &brg_min, &brg_max);

		for (brg = brg_min; brg <= brg_max; brg++)
		{
			tmp		= (float) hdlc->osc.prog / ctr0 /  ctr1 / dpll / (brg + 1);
			diff	= tmp - bit_rate;
			diff	= ABS(diff);

			if (diff < diff_best)
			{
				diff_best	= diff;
				rate		= tmp;

				if (option->brg0->clk_src >= 0)
					hdlc->usc.brg0.divider	= brg;
				else
					hdlc->usc.brg1.divider	= brg;
			}
		}
	}

	return(rate);
}



//*****************************************************************************
float hdlc_cfg_rxc_ctr_brg_dpll_rxclk(sio4_hdlc_t* hdlc, s32 bit_rate)
{
	float		diff;
	float		diff_best	= _FLOAT_MAX_;
	int			i;
	float		rate		= (float) hdlc->osc.prog;
	float		rate_best	= _FLOAT_MAX_;
	int			reuse;
	sio4_hdlc_t	tmp;
	sio4_hdlc_t	tmp_best	= hdlc[0];
	int			use;

	for (reuse = 1; reuse >= 0; reuse--)
	{
		for (i = 0; i < (int) SIZEOF_ARRAY(_options); i++)
		{
			tmp	= hdlc[0];
			_apply_option(&tmp, _options + i);
			use	= _usc_check_usage(hdlc, &tmp, _options + i, reuse);

			if (use)
			{
				rate	= _compute_rx_rate(bit_rate, hdlc, &tmp, _options + i);
				diff	= rate - bit_rate;
				diff	= ABS(diff);

				if (diff < diff_best)
				{
					diff_best	= diff;
					tmp_best	= tmp;
					rate_best	= rate;
				}
			}
		}
	}

	if (diff_best < _FLOAT_MAX_)
		hdlc[0]	= tmp_best;

	return(rate_best);
}


